@@ -34,11 +34,19 @@
 #include <glib.h>
 #include <locale.h>
 
+#ifdef ENABLE_GNOME_KEYRING
+#include <gnome-keyring.h>
+#endif
+
 #include "defines.h"
 #include "../config.h"
 
 #include "gksu-context.h"
 
+#ifdef ENABLE_GNOME_KEYRING
+gboolean keyring_used = FALSE;
+#endif
+
 /* local function declarations */
 static void
 gksu_context_init (GTypeInstance *instance, gpointer g_class);
@@ -52,6 +60,14 @@ gksu_context_finalize (GObject *object);
 static void
 gksu_context_class_init (GksuContextClass *klass);
 
+void
+nullify_password (GksuContext *context)
+{
+  if (context->password)
+    g_free (context->password);
+  context->password = NULL;
+}
+
 /**
  * gksu_context_new:
  *
@@ -585,6 +601,189 @@ sudo_reset_xauth (GksuContext *context, gchar *xauth,
   g_free (xauth);
 }
 
+#ifdef ENABLE_GNOME_KEYRING
+gboolean
+try_gnome_keyring_password (GksuContext *context)
+{
+  GnomeKeyringAttributeList *attributes;
+  GnomeKeyringAttribute attribute;
+  GnomeKeyringResult result;
+  GList *list;
+  GError *error = NULL;
+  gboolean keyring_has_password;
+  gchar *keyring_password;
+  gchar *keyring_command;
+
+  keyring_used = TRUE;
+
+  attributes = gnome_keyring_attribute_list_new ();
+
+  attribute.name = g_strdup ("user");
+  attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
+  attribute.value.string = g_strdup (gksu_context_get_user (context));
+  g_array_append_val (attributes, attribute);
+
+  attribute.name = g_strdup ("type");
+  attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
+  attribute.value.string = g_strdup ("local");
+  g_array_append_val (attributes, attribute);
+
+  attribute.name = g_strdup ("creator");
+  attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
+  attribute.value.string = g_strdup ("gksu");
+  g_array_append_val (attributes, attribute);
+      
+  list = g_list_alloc();
+  keyring_has_password = FALSE;
+
+  result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_GENERIC_SECRET,
+					  attributes,
+					  &list);
+  gnome_keyring_attribute_list_free (attributes);
+  if (
+      (result == GNOME_KEYRING_RESULT_OK) &&
+      (g_list_length(list) == 1)
+      )
+    {
+      GnomeKeyringFound *found = list->data;
+      keyring_password = g_strdup(found->secret);
+
+      keyring_command = g_strdup (gksu_context_get_command (context));
+      gksu_context_set_command (context, "/bin/echo test > /dev/null");
+      gksu_context_set_password (context, keyring_password);
+      gksu_context_run (context, &error);
+
+      if (!error)
+	keyring_has_password = TRUE;
+      gksu_context_set_command (context, keyring_command);
+    }
+
+  if (keyring_has_password) 
+    {
+      gchar *password;
+
+      password = g_locale_from_utf8 (keyring_password, 
+				     strlen (keyring_password), 
+				     NULL, NULL, NULL);
+      gksu_context_set_password (context, password);
+      return FALSE; /* we do not need a password */
+    }
+
+  return TRUE;
+}
+#endif
+
+gboolean
+try_su_run (GksuContext *context)
+{
+  int fdpty;
+  pid_t pid;
+  int i = 0;
+
+  gboolean need_pass = TRUE;
+
+  pid = forkpty (&fdpty, NULL, NULL, NULL);
+  if (pid == 0)
+    {
+      gchar **cmd = g_malloc (sizeof(gchar*)*7);
+
+      cmd[i] = g_strdup ("/bin/su"); i++;
+      if (context->login_shell)
+	{
+	  cmd[i] = g_strdup ("-"); i++;
+	}
+      cmd[i] = g_strdup (context->user); i++;
+      if (context->keep_env)
+	{
+	  cmd[i] = g_strdup ("-p"); i++;
+	}
+      cmd[i] = g_strdup ("-c"); i++;
+
+      /* needs to get X authorization prior to running the program */
+      cmd[i] = g_strdup_printf ("echo > /dev/null"); i++;
+
+      cmd[i] = NULL;
+
+      /* executes the command */
+      if (execv (cmd[0], cmd) == -1)
+	{
+	  fprintf (stderr, 
+		   _("Unable to run /bin/su: %s"),
+		   strerror(errno));
+
+	  for (i = 0 ; cmd[i] != NULL ; i++)
+	    g_free (cmd[i]);
+	  g_free(cmd);
+	  
+	  return TRUE;
+	}
+    }
+  else if (pid == -1)
+    {
+      return TRUE;
+    }
+  else
+    {
+      fd_set rfds;
+
+      struct timeval tv;
+
+      gchar buf[256];
+      gint status;
+
+      FD_ZERO(&rfds);
+      FD_SET(fdpty, &rfds);
+
+      tv.tv_sec = 0;
+      tv.tv_usec = 100;
+
+      if (select (fdpty+1, &rfds, NULL, NULL, &tv) < 0)
+	return TRUE;
+
+      read (fdpty, buf, 256);
+      if (context->debug)
+	fprintf (stderr, "try_su_run: buf: -%s-\n", buf);
+
+      if (!strncmp (buf, "Password: ", 10))
+	{
+	  write (fdpty, "\n", 1);
+	  close (fdpty);
+	}
+      else
+	need_pass = FALSE;
+
+      waitpid (pid, &status, 0);
+    }
+
+  return need_pass;
+}
+
+/**
+ * gksu_context_try_need_password
+ * @context: a #GksuContext
+ *
+ * Checks if we need to ask for a password or if we have ways of
+ * getting the password for ourselves or we simply don't need it.
+ *
+ * Returns: TRUE if requesting a password is needed, FALSE otherwise.
+ *
+ */
+gboolean
+gksu_context_try_need_password (GksuContext *context)
+{
+  nullify_password (context);
+
+  if (try_su_run (context) == FALSE)
+    return FALSE;
+
+#ifdef ENABLE_GNOME_KEYRING
+  if (try_gnome_keyring_password (context) == FALSE)
+    return FALSE;
+#endif
+
+  return TRUE;
+}
+
 /**
  * gksu_context_run:
  * @context: a #GksuContext
@@ -598,9 +797,9 @@ sudo_reset_xauth (GksuContext *context, gchar *xauth,
 gboolean
 gksu_context_run (GksuContext *context, GError **error)
 {
+  GQuark gksu_quark;
   int i = 0;
   gboolean auth_failed = FALSE;
-  GQuark gksu_quark;
 
   gchar auxcommand[] = PREFIX "/lib/" PACKAGE "/gksu-run-helper";
 
@@ -692,10 +891,6 @@ gksu_context_run (GksuContext *context, GError **error)
       */
       gboolean first_line_read = FALSE;
       
-      read (fdpty, buf, 256);
-      if (context->debug)
-	fprintf (stderr, "gksu_context_run: buf: -%s-\n", buf);
-      
       my_uid = getuid();
       pwd = getpwnam (context->user);
       if (pwd)
@@ -704,19 +899,16 @@ gksu_context_run (GksuContext *context, GError **error)
       /* no need to ask for password if we're already root */
       if (my_uid != target_uid && my_uid)
 	{
-	  if (context->password == NULL)
+	  if (context->password)
 	    {
-	      /* FIXME: need to check if we can run the program
-		 without using a password (pam_wheel?) and if
-		 not, return a comprehensive error message/code */
-	      write (fdpty, "\n", 1);
-	      return -1;
+	      read (fdpty, buf, 256);
+	      if (context->debug)
+		fprintf (stderr, "gksu_context_run: buf: -%s-\n", buf);
+	      sleep (1);
+	      write (fdpty, context->password, 
+		     strlen(context->password));
 	    }
-
-	  sleep (1);
-	  write (fdpty, context->password, 
-		 strlen(context->password));
-	} 
+	}
       
       FD_ZERO(&rfds);
       FD_SET(fdpty, &rfds);
@@ -742,7 +934,12 @@ gksu_context_run (GksuContext *context, GError **error)
 
 	      if (!first_line_read)
 		{
-		  read (fdpty, buf, 255);
+		  /* we've already read something important on the
+		     outside of the loop, probably because su gave
+		     no output at all (pam_wheel trust or something)
+		  */
+		  if (strncmp (buf, "gksu", 4))
+		    read (fdpty, buf, 255);
 
 		  if (!strncmp (buf, "su: Authentication failure", 26))
 		    auth_failed = TRUE;
@@ -761,6 +958,56 @@ gksu_context_run (GksuContext *context, GError **error)
 		      bzero (buf, 255);
 		    }
 
+#ifdef ENABLE_GNOME_KEYRING
+		  if (keyring_used && context->password)
+		    {
+		      static GMainLoop *keyring_loop = NULL;
+		      GnomeKeyringAttributeList *attributes;
+		      GnomeKeyringAttribute attribute;
+		      
+		      gchar *key_name;
+		      
+		      attributes = gnome_keyring_attribute_list_new ();
+		      
+		      attribute.name = g_strdup ("user");
+		      attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
+		      attribute.value.string = g_strdup (gksu_context_get_user (context));
+		      g_array_append_val (attributes, attribute);
+		    
+		      attribute.name = g_strdup ("type");
+		      attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
+		      attribute.value.string = g_strdup ("local");
+		      g_array_append_val (attributes, attribute);
+
+		      attribute.name = g_strdup ("creator");
+		      attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
+		      attribute.value.string = g_strdup ("gksu");
+		      g_array_append_val (attributes, attribute);
+      
+		      key_name = g_strdup_printf ("Local password for user %s",
+						  gksu_context_get_user (context));
+      
+		      keyring_loop = g_main_loop_new (NULL, FALSE);
+
+		      void
+			keyring_create_item_cb (GnomeKeyringResult result,
+						guint32 id, gpointer data)
+			{
+			  g_main_loop_quit (keyring_loop);
+			}
+
+		      gnome_keyring_item_create (NULL,
+						 GNOME_KEYRING_ITEM_GENERIC_SECRET,
+						 key_name,
+						 attributes,
+						 gksu_context_get_password (context),
+						 TRUE,
+						 keyring_create_item_cb, NULL, NULL);
+		      gnome_keyring_attribute_list_free (attributes);
+		      g_main_loop_run (keyring_loop);
+		    }
+#endif
+
 		  first_line_read = TRUE;
 		}
 	      else if (!strncmp (buf, "gksu-run:", 8))
@@ -814,6 +1061,151 @@ gksu_context_run (GksuContext *context, GError **error)
   return 0;
 }
 
+static gboolean
+try_sudo_validation (GksuContext *context)
+{
+  char **cmd;
+  int argcount = 4;
+  char buffer[256];
+
+  pid_t pid;
+  int status;
+  size_t r;
+  FILE *infile, *outfile;
+  int parent_pipe[2];	/* For talking to the parent */
+  int child_pipe[2];	/* For talking to the child */
+
+  gboolean need_pass = TRUE;
+  
+  bzero(buffer, 256);
+
+  if ((pipe(parent_pipe)) == -1)
+    return TRUE;
+
+  if ((pipe(child_pipe)) == -1)
+    return TRUE;
+
+  cmd = g_new (gchar *, argcount + 1);
+  
+  argcount = 0;
+
+  /* sudo binary */
+  cmd[argcount] = g_strdup("/usr/bin/sudo");
+  argcount++;
+
+  /* Make sudo read from STDIN */
+  cmd[argcount] = g_strdup("-S");
+  argcount++;
+
+  /* Make sudo use next arg as prompt */
+  cmd[argcount] = g_strdup("-p");
+  argcount++;
+
+  /* prompt */
+  cmd[argcount] = g_strdup("GNOME_SUDO_PASS");
+  argcount++;
+
+  /* just try to validate the password */
+  cmd[argcount] = g_strdup("-v");
+  argcount++;
+
+  cmd[argcount] = NULL;
+
+  pid = fork();
+  if (pid == -1)
+    return TRUE;
+  else if (pid == 0) 
+    {
+      // Child
+      close(child_pipe[1]);
+      dup2(child_pipe[0], STDIN_FILENO);
+      dup2(parent_pipe[1], STDERR_FILENO);
+      
+      execv(cmd[0], cmd);
+      
+      return TRUE;
+    } 
+  else 
+    {
+      // Parent
+      close(parent_pipe[1]);
+      
+      infile = fdopen(parent_pipe[0], "r");
+      if (!infile)
+	return TRUE;
+
+      outfile = fdopen(child_pipe[1], "w");
+      if (!outfile)
+	return TRUE;
+
+      /*
+	we are expecting to receive a GNOME_SUDO_PASS
+	if we don't there are two possibilities: an error
+	or a password is not needed
+      */
+      r = read(parent_pipe[0], buffer, 15);
+      buffer[16] = '\0';
+
+      if (context->debug)
+	fprintf (stderr, "buffer: -%s-\n", buffer);
+
+      if (strncmp(buffer, "GNOME_SUDO_PASS", 15) == 0) 
+	{
+	  if (context->debug)
+	    fprintf (stderr, "We'll need a password...\n");
+
+	  fprintf (outfile, "\n");
+	  fclose (outfile);
+	} 
+      else 
+	{
+	  if (context->debug)
+	    fprintf (stderr, "We won't need a password, it seems!\n");
+	  need_pass = FALSE;
+	}
+
+
+      while (!waitpid (pid, &status, WNOHANG))
+	{
+	  bzero(buffer, 256);
+	  if(!fgets (buffer, 255, infile))
+	    break;
+	  fprintf (stderr, "%s", buffer);
+	}
+
+      /* make sure we did read everything */
+      while (1)
+	{
+	  bzero(buffer, 256);
+	  if(!fread (buffer, sizeof(gchar), 255, infile))
+	    break;
+	  fprintf (stderr, "%s", buffer);
+	  fflush (stderr);
+	}
+    }
+
+  return need_pass;
+}
+
+/**
+ * gksu_context_sudo_try_need_password
+ * @context: a #GksuContext
+ *
+ * Checks if we need to ask for a password or if we have ways of
+ * getting the password for ourselves or we simply don't need it.
+ *
+ * Returns: TRUE if requesting a password is needed, FALSE otherwise.
+ *
+ */
+gboolean
+gksu_context_sudo_try_need_password (GksuContext *context)
+{
+  if (!try_sudo_validation (context))
+    return FALSE;
+
+  return TRUE;
+}
+
 /**
  * gksu_context_sudo_run:
  * @context: a #GksuContext
@@ -980,6 +1372,7 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
   else if (pid == 0) 
     {
       // Child
+      setsid();   // make us session leader
       close(child_pipe[1]);
       dup2(child_pipe[0], STDIN_FILENO);
       dup2(parent_pipe[1], STDERR_FILENO);
